"""
    Tests for the NURBS-Python package
    Released under The MIT License. See LICENSE file for details.
    Copyright (c) 2018-2019 Onur Rauf Bingol

    Requires "pytest" to run.
"""

from nurbs import BSpline, convert, evaluators, helpers, linalg, operations
from pytest import fixture, mark

GEOMDL_DELTA = 0.001


@fixture
def spline_curve():
    """Creates a spline Curve"""
    curve = BSpline.Curve()
    curve.degree = 3
    curve.ctrlpts = [[5.0, 5.0], [10.0, 10.0], [20.0, 15.0], [35.0, 15.0], [45.0, 10.0], [50.0, 5.0]]
    curve.knotvector = [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]
    return curve


def test_bspline_curve_name(spline_curve):
    spline_curve.name = "Curve Testing"
    assert spline_curve.name == "Curve Testing"


def test_bspline_curve_degree(spline_curve):
    assert spline_curve.degree == 3


def test_bspline_curve_ctrlpts(spline_curve):
    assert spline_curve.ctrlpts == [[5.0, 5.0], [10.0, 10.0], [20.0, 15.0], [35.0, 15.0], [45.0, 10.0], [50.0, 5.0]]
    assert spline_curve.dimension == 2


def test_bspline_curve_knot_vector(spline_curve):
    assert spline_curve.knotvector == [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]


@mark.parametrize(
    "param, res",
    [(0.0, (5.0, 5.0)), (0.3, (18.617, 13.377)), (0.5, (27.645, 14.691)), (0.6, (32.143, 14.328)), (1.0, (50.0, 5.0))],
)
def test_bspline_curve2d_eval(spline_curve, param, res):
    evalpt = spline_curve.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA


def test_bspline_curve2d_deriv(spline_curve):
    der1 = spline_curve.derivatives(u=0.35, order=2)
    spline_curve.evaluator = evaluators.CurveEvaluator2()
    der2 = spline_curve.derivatives(u=0.35, order=2)

    assert abs(der1[0][0] - der2[0][0]) < GEOMDL_DELTA
    assert abs(der1[0][1] - der2[0][1]) < GEOMDL_DELTA
    assert abs(der1[1][0] - der2[1][0]) < GEOMDL_DELTA
    assert abs(der1[1][1] - der2[1][1]) < GEOMDL_DELTA
    assert abs(der1[2][0] - der2[2][0]) < GEOMDL_DELTA
    assert abs(der1[2][1] - der2[2][1]) < GEOMDL_DELTA


def test_bspline_curve2d_deriv_eval(spline_curve):
    evalpt = spline_curve.evaluate_single(0.35)
    der1 = spline_curve.derivatives(u=0.35)
    spline_curve.evaluator = evaluators.CurveEvaluator2()
    der2 = spline_curve.derivatives(u=0.35)

    assert abs(der1[0][0] - evalpt[0]) < GEOMDL_DELTA
    assert abs(der1[0][1] - evalpt[1]) < GEOMDL_DELTA
    assert abs(der2[0][0] - evalpt[0]) < GEOMDL_DELTA
    assert abs(der2[0][1] - evalpt[1]) < GEOMDL_DELTA


@mark.parametrize(
    "param, num_insert, res", [(0.3, 1, (18.617, 13.377)), (0.6, 1, (32.143, 14.328)), (0.6, 2, (32.143, 14.328))]
)
def test_bspline_curve2d_insert_knot(spline_curve, param, num_insert, res):
    s_pre = helpers.find_multiplicity(param, spline_curve.knotvector)
    spline_curve.insert_knot(param, num=num_insert)
    s_post = helpers.find_multiplicity(param, spline_curve.knotvector)
    evalpt = spline_curve.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA
    assert s_pre + num_insert == s_post


def test_bspline_curve2d_insert_knot_kv(spline_curve):
    spline_curve.insert_knot(0.66, num=2)
    s = helpers.find_multiplicity(0.66, spline_curve.knotvector)

    assert spline_curve.knotvector[5] == 0.66
    assert s == 3


@mark.parametrize("param, num_remove", [(0.33, 1), (0.66, 1)])
def test_bspline_curve2d_remove_knot(spline_curve, param, num_remove):
    s_pre = helpers.find_multiplicity(param, spline_curve.knotvector)
    c_pre = spline_curve.ctrlpts_size
    spline_curve.remove_knot(param, num=num_remove)
    s_post = helpers.find_multiplicity(param, spline_curve.knotvector)
    c_post = spline_curve.ctrlpts_size

    assert c_pre - num_remove == c_post
    assert s_pre - num_remove == s_post


def test_bspline_curve2d_remove_knot_kv(spline_curve):
    spline_curve.remove_knot(0.66, num=1)
    s = helpers.find_multiplicity(0.66, spline_curve.knotvector)

    assert 0.66 not in spline_curve.knotvector
    assert s == 0


@mark.parametrize(
    "density, kv",
    [
        (0, [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]),
        (
            1,
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.165,
                0.165,
                0.165,
                0.33,
                0.33,
                0.33,
                0.495,
                0.495,
                0.495,
                0.66,
                0.66,
                0.66,
                0.830,
                0.830,
                0.830,
                1.0,
                1.0,
                1.0,
                1.0,
            ],
        ),
    ],
)
def test_bspline_curve2d_knot_refine(spline_curve, density, kv):
    operations.refine_knotvector(spline_curve, [density])
    for a, b in zip(kv, spline_curve.knotvector):
        assert abs(a - b) < GEOMDL_DELTA


def test_bspline_curve2d_degree_elevate_degree(spline_curve):
    dops = 1
    degree_new = spline_curve.degree + dops
    operations.degree_operations(spline_curve, [dops])
    assert spline_curve.degree == degree_new


def test_bspline_curve2d_degree_elevate_ctrlpts_size(spline_curve):
    dops = 1
    ctrlpts_size = spline_curve.ctrlpts_size + dops
    operations.degree_operations(spline_curve, [dops])
    assert spline_curve.ctrlpts_size == ctrlpts_size


def test_bspline_curve2d_degree_reduce_degree(spline_curve):
    dops = -1
    degree_new = spline_curve.degree + dops
    operations.degree_operations(spline_curve, [dops])
    assert spline_curve.degree == degree_new


def test_bspline_curve2d_degree_reduce_ctrlpts_size(spline_curve):
    dops = -1
    ctrlpts_size_new = spline_curve.ctrlpts_size + dops
    operations.degree_operations(spline_curve, [dops])
    assert spline_curve.ctrlpts_size == ctrlpts_size_new


@fixture
def spline_curve3d(spline_curve):
    curve3d = operations.add_dimension(spline_curve, offset=1.0)
    return curve3d


def test_bspline_curve3d_ctrlpts(spline_curve3d):
    assert spline_curve3d.ctrlpts == [
        [5.0, 5.0, 1.0],
        [10.0, 10.0, 1.0],
        [20.0, 15.0, 1.0],
        [35.0, 15.0, 1.0],
        [45.0, 10.0, 1.0],
        [50.0, 5.0, 1.0],
    ]
    assert spline_curve3d.dimension == 3


@fixture
def nurbs_curve(spline_curve):
    curve = convert.bspline_to_nurbs(spline_curve)
    curve.weights = [0.5, 1.0, 0.75, 1.0, 0.25, 1.0]
    return curve


def test_nurbs_curve2d_weights(nurbs_curve):
    assert nurbs_curve.weights == [0.5, 1.0, 0.75, 1.0, 0.25, 1.0]


@mark.parametrize(
    "param, res", [(0.0, (5.0, 5.0)), (0.2, (13.8181, 11.5103)), (0.5, (28.1775, 14.7858)), (0.95, (48.7837, 6.0022))]
)
def test_nurbs_curve2d_eval(nurbs_curve, param, res):
    evalpt = nurbs_curve.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA


@mark.parametrize(
    "param, order, res",
    [
        (0.0, 1, ((5.0, 5.0), (90.9090, 90.9090))),
        (0.2, 2, ((13.8181, 11.5103), (40.0602, 17.3878), (104.4062, -29.3672))),
        (0.5, 3, ((28.1775, 14.7858), (39.7272, 2.2562), (-116.9254, -49.7367), (125.5276, 196.8865))),
        (0.95, 1, ((48.7837, 6.0022), (39.5178, -29.9962))),
    ],
)
def test_nurbs_curve2d_deriv(nurbs_curve, param, order, res):
    deriv = nurbs_curve.derivatives(u=param, order=order)

    for computed, expected in zip(deriv, res):
        for c, e in zip(computed, expected):
            assert abs(c - e) < GEOMDL_DELTA


@fixture
def spline_curve_kv_norm1():
    """Creates a spline Curve with knot vector normalization"""
    curve = BSpline.Curve(normalize_kv=True)
    curve.degree = 3
    curve.ctrlpts = [[5.0, 5.0], [10.0, 10.0], [20.0, 15.0], [35.0, 15.0], [45.0, 10.0], [50.0, 5.0]]
    curve.knotvector = [0.0, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0, 3.0, 3.0, 3.0]
    return curve


@mark.parametrize(
    "param, res",
    [
        (0.0, (5.0, 5.0)),
        (0.33333, (19.9998, 13.7499)),
        (0.5, (27.5, 14.687)),
        (0.66666, (34.9997, 13.75)),
        (1.0, (50.0, 5.0)),
    ],
)
def test_bspline_curve2d_eval_kv_norm1(spline_curve_kv_norm1, param, res):
    evalpt = spline_curve_kv_norm1.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA


@fixture
def spline_curve_kv_norm2():
    """Creates a spline Curve without knot vector normalization"""
    curve = BSpline.Curve(normalize_kv=False)
    curve.degree = 3
    curve.ctrlpts = [[5.0, 5.0], [10.0, 10.0], [20.0, 15.0], [35.0, 15.0], [45.0, 10.0], [50.0, 5.0]]
    curve.knotvector = [0.0, 0.0, 0.0, 0.0, 1.0, 2.0, 3.0, 3.0, 3.0, 3.0]
    return curve


@mark.parametrize(
    "param, res",
    [(0.0, (5.0, 5.0)), (1.0, (19.9998, 13.7499)), (1.5, (27.5, 14.687)), (2.0, (34.9997, 13.75)), (3.0, (50.0, 5.0))],
)
def test_bspline_curve2d_eval_kv_norm2(spline_curve_kv_norm2, param, res):
    evalpt = spline_curve_kv_norm2.evaluate_single(param)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA



"""
    Tests for the NURBS-Python package
    Released under The MIT License. See LICENSE file for details.
    Copyright (c) 2018-2019 Onur Rauf Bingol

    Requires "pytest" to run.
"""

from nurbs import BSpline, convert, evaluators, helpers
from pytest import fixture, mark

GEOMDL_DELTA = 0.001


@fixture
def spline_surf():
    """Creates a B-spline surface instance"""
    # Create a surface instance
    surf = BSpline.Surface()

    # Set degrees
    surf.degree_u = 3
    surf.degree_v = 3

    ctrlpts = [
        [-25.0, -25.0, -10.0],
        [-25.0, -15.0, -5.0],
        [-25.0, -5.0, 0.0],
        [-25.0, 5.0, 0.0],
        [-25.0, 15.0, -5.0],
        [-25.0, 25.0, -10.0],
        [-15.0, -25.0, -8.0],
        [-15.0, -15.0, -4.0],
        [-15.0, -5.0, -4.0],
        [-15.0, 5.0, -4.0],
        [-15.0, 15.0, -4.0],
        [-15.0, 25.0, -8.0],
        [-5.0, -25.0, -5.0],
        [-5.0, -15.0, -3.0],
        [-5.0, -5.0, -8.0],
        [-5.0, 5.0, -8.0],
        [-5.0, 15.0, -3.0],
        [-5.0, 25.0, -5.0],
        [5.0, -25.0, -3.0],
        [5.0, -15.0, -2.0],
        [5.0, -5.0, -8.0],
        [5.0, 5.0, -8.0],
        [5.0, 15.0, -2.0],
        [5.0, 25.0, -3.0],
        [15.0, -25.0, -8.0],
        [15.0, -15.0, -4.0],
        [15.0, -5.0, -4.0],
        [15.0, 5.0, -4.0],
        [15.0, 15.0, -4.0],
        [15.0, 25.0, -8.0],
        [25.0, -25.0, -10.0],
        [25.0, -15.0, -5.0],
        [25.0, -5.0, 2.0],
        [25.0, 5.0, 2.0],
        [25.0, 15.0, -5.0],
        [25.0, 25.0, -10.0],
    ]

    # Set control points
    surf.set_ctrlpts(ctrlpts, 6, 6)

    # Set knot vectors
    surf.knotvector_u = [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]
    surf.knotvector_v = [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]

    return surf


def test_bspline_curve_name(spline_surf):
    spline_surf.name = "Surface Testing"
    assert spline_surf.name == "Surface Testing"


def test_bspline_surface_degree_u(spline_surf):
    assert spline_surf.degree_u == 3


def test_bspline_surface_degree_v(spline_surf):
    assert spline_surf.degree_v == 3


def test_bspline_surface_ctrlpts(spline_surf):
    assert spline_surf.ctrlpts2d[1][1] == [-15.0, -15.0, -4.0]
    assert spline_surf.dimension == 3


def test_bspline_surface_knot_vector_u(spline_surf):
    assert spline_surf.knotvector_u == [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]


def test_bspline_surface_knot_vector_v(spline_surf):
    assert spline_surf.knotvector_v == [0.0, 0.0, 0.0, 0.0, 0.33, 0.66, 1.0, 1.0, 1.0, 1.0]


@mark.parametrize(
    "param, res",
    [
        ((0.0, 0.0), (-25.0, -25.0, -10.0)),
        ((0.0, 0.2), (-25.0, -11.403, -3.385)),
        ((0.0, 1.0), (-25.0, 25.0, -10.0)),
        ((0.3, 0.0), (-7.006, -25.0, -5.725)),
        ((0.3, 0.4), [-7.006, -3.308, -6.265]),
        ((0.3, 1.0), [-7.006, 25.0, -5.725]),
        ((0.6, 0.0), (3.533, -25.0, -4.224)),
        ((0.6, 0.6), (3.533, 3.533, -6.801)),
        ((0.6, 1.0), (3.533, 25.0, -4.224)),
        ((1.0, 0.0), (25.0, -25.0, -10.0)),
        ((1.0, 0.8), (25.0, 11.636, -2.751)),
        ((1.0, 1.0), (25.0, 25.0, -10.0)),
    ],
)
def test_bspline_surface_eval(spline_surf, param, res):
    evalpt = spline_surf.evaluate_single(param)
    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA
    assert abs(evalpt[2] - res[2]) < GEOMDL_DELTA


def test_bspline_surface_deriv(spline_surf):
    der1 = spline_surf.derivatives(u=0.35, v=0.35, order=2)
    spline_surf.evaluator = evaluators.SurfaceEvaluator2()
    der2 = spline_surf.derivatives(u=0.35, v=0.35, order=2)
    for k in range(0, 3):
        for l in range(0, 3 - k):
            assert abs(der1[k][l][0] - der2[k][l][0]) < GEOMDL_DELTA
            assert abs(der1[k][l][1] - der2[k][l][1]) < GEOMDL_DELTA
            assert abs(der1[k][l][2] - der2[k][l][2]) < GEOMDL_DELTA


@mark.parametrize(
    "params, uv, res",
    [
        (dict(u=0.3, v=0.4), (0.3, 0.4), (-7.006, -3.308, -6.265)),
        (dict(u=0.3, num_u=2), (0.3, 0.4), (-7.006, -3.308, -6.265)),
        (dict(v=0.3, num_v=2), (0.3, 0.4), (-7.006, -3.308, -6.265)),
    ],
)
def test_bspline_surface_insert_knot_eval(spline_surf, params, uv, res):
    # Insert knot
    spline_surf.insert_knot(**params)

    # Evaluate surface
    evalpt = spline_surf.evaluate_single(uv)

    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA
    assert abs(evalpt[2] - res[2]) < GEOMDL_DELTA


@mark.parametrize("params, idx, val", [(dict(v=0.3, num_v=2), 4, 0.3), (dict(v=0.3, num_v=2), 6, 0.33)])
def test_bspline_surface_insert_knot_kv_v(spline_surf, params, idx, val):
    # Insert knot
    spline_surf.insert_knot(**params)

    assert spline_surf.knotvector_v[idx] == val


@mark.parametrize("params, idx, val", [(dict(u=0.33, num_u=2), 3, 0.0), (dict(u=0.33, num_u=1), 6, 0.66)])
def test_bspline_surface_insert_kv_u(spline_surf, params, idx, val):
    # Insert knot
    spline_surf.insert_knot(**params)

    assert spline_surf.knotvector_u[idx] == val


@mark.parametrize("param, num_remove", [(0.33, 1), (0.66, 1)])
def test_bspline_surface_remove_knot_u(spline_surf, param, num_remove):
    s_pre = helpers.find_multiplicity(param, spline_surf.knotvector_u)
    c_pre = spline_surf.ctrlpts_size_u
    spline_surf.remove_knot(u=param, num_u=num_remove)
    s_post = helpers.find_multiplicity(param, spline_surf.knotvector_u)
    c_post = spline_surf.ctrlpts_size_u

    assert c_pre - num_remove == c_post
    assert s_pre - num_remove == s_post


@mark.parametrize("param, num_remove", [(0.33, 1), (0.66, 1)])
def test_bspline_surface_remove_knot_v(spline_surf, param, num_remove):
    s_pre = helpers.find_multiplicity(param, spline_surf.knotvector_v)
    c_pre = spline_surf.ctrlpts_size_v
    spline_surf.remove_knot(v=param, num_v=num_remove)
    s_post = helpers.find_multiplicity(param, spline_surf.knotvector_v)
    c_post = spline_surf.ctrlpts_size_v

    assert c_pre - num_remove == c_post
    assert s_pre - num_remove == s_post


def test_bspline_surface_remove_knot_kv_u(spline_surf):
    spline_surf.remove_knot(u=0.66, num_u=1)
    s = helpers.find_multiplicity(0.66, spline_surf.knotvector_u)

    assert 0.66 not in spline_surf.knotvector_u
    assert s == 0


def test_bspline_surface_remove_knot_kv_v(spline_surf):
    spline_surf.remove_knot(v=0.33, num_v=1)
    s = helpers.find_multiplicity(0.33, spline_surf.knotvector_v)

    assert 0.33 not in spline_surf.knotvector_v
    assert s == 0


@fixture
def nurbs_surf(spline_surf):
    surf = convert.bspline_to_nurbs(spline_surf)
    return surf


def test_nurbs_weights(nurbs_surf):
    assert len(nurbs_surf.weights) == nurbs_surf.ctrlpts_size
    assert nurbs_surf.weights[5] == 1.0


@mark.parametrize(
    "param, res",
    [
        ((0.0, 0.0), (-25.0, -25.0, -10.0)),
        ((0.0, 0.2), (-25.0, -11.403, -3.385)),
        ((0.0, 1.0), (-25.0, 25.0, -10.0)),
        ((0.3, 0.0), (-7.006, -25.0, -5.725)),
        ((0.3, 0.4), [-7.006, -3.308, -6.265]),
        ((0.3, 1.0), [-7.006, 25.0, -5.725]),
        ((0.6, 0.0), (3.533, -25.0, -4.224)),
        ((0.6, 0.6), (3.533, 3.533, -6.801)),
        ((0.6, 1.0), (3.533, 25.0, -4.224)),
        ((1.0, 0.0), (25.0, -25.0, -10.0)),
        ((1.0, 0.8), (25.0, 11.636, -2.751)),
        ((1.0, 1.0), (25.0, 25.0, -10.0)),
    ],
)
def test_nurbs_surface_eval(nurbs_surf, param, res):
    evalpt = nurbs_surf.evaluate_single(param)
    assert abs(evalpt[0] - res[0]) < GEOMDL_DELTA
    assert abs(evalpt[1] - res[1]) < GEOMDL_DELTA
    assert abs(evalpt[2] - res[2]) < GEOMDL_DELTA


@mark.parametrize(
    "param, order, res",
    [
        (
            (0.0, 0.25),
            1,
            [
                [[-25.0, -9.0771, -2.3972], [5.5511e-15, 43.6910, 17.5411]],
                [[90.9090, 0.0, -15.0882], [-5.9750e-15, 0.0, -140.0367]],
            ],
        ),
        (
            (0.95, 0.75),
            2,
            [
                [[20.8948, 9.3097, -2.4845], [-1.1347e-14, 43.7672, -15.0153], [-5.0393e-30, 100.1022, -74.1165]],
                [
                    [76.2308, -1.6965e-15, 18.0372],
                    [9.8212e-15, -5.9448e-15, -158.5462],
                    [4.3615e-30, -2.4356e-13, -284.3037],
                ],
                [
                    [224.5342, -5.6794e-14, 93.3843],
                    [4.9856e-14, -4.0400e-13, -542.6274],
                    [2.2140e-29, -1.88662e-12, -318.8808],
                ],
            ],
        ),
    ],
)
def test_nurbs_surface_deriv(nurbs_surf, param, order, res):
    deriv = nurbs_surf.derivatives(*param, order=order)

    for computed, expected in zip(deriv, res):
        for idx in range(order + 1):
            for c, e in zip(computed[idx], expected[idx]):
                assert abs(c - e) < GEOMDL_DELTA


def test_surface_bounding_box(spline_surf):
    # Evaluate bounding box
    to_check = spline_surf.bbox

    # Evaluation result
    result = ((-25.0, -25.0, -10.0), (25.0, 25.0, 2.0))

    assert abs(to_check[0][0] - result[0][0]) < GEOMDL_DELTA
    assert abs(to_check[0][1] - result[0][1]) < GEOMDL_DELTA
    assert abs(to_check[0][2] - result[0][2]) < GEOMDL_DELTA
    assert abs(to_check[1][0] - result[1][0]) < GEOMDL_DELTA
    assert abs(to_check[1][1] - result[1][1]) < GEOMDL_DELTA
    assert abs(to_check[1][2] - result[1][2]) < GEOMDL_DELTA



"""
.. module:: _voxelize
    :platform: Unix, Windows
    :synopsis: Helper functions for voxelization module

.. moduleauthor:: Onur Rauf Bingol <orbingol@gmail.com>

"""

from functools import partial

from . import linalg
from ._utilities import pool_context
from .exceptions import GeomdlException

# Initialize an empty __all__ for controlling imports
__all__ = []


def find_inouts_st(voxel_grid, datapts, **kwargs):
    """Single-threaded ins and outs finding (default)

    :param voxel_grid: voxel grid
    :param datapts: data points
    :return: in-outs
    """
    tol = kwargs.get("tol", 10e-8)
    filled = [0 for _ in range(len(voxel_grid))]
    for idx, bb in enumerate(voxel_grid):
        pts_inside = is_point_inside_voxel(bb, datapts, tol=tol)
        if pts_inside:
            filled[idx] = 1
    return filled


def find_inouts_mp(voxel_grid, datapts, **kwargs):
    """Multi-threaded ins and outs finding (using multiprocessing)

    :param voxel_grid: voxel grid
    :param datapts: data points
    :return: in-outs
    """
    tol = kwargs.get("tol", 10e-8)
    num_procs = kwargs.get("num_procs", 4)
    with pool_context(processes=num_procs) as pool:
        filled = pool.map(partial(is_point_inside_voxel, ptsarr=datapts, tol=tol), voxel_grid)
    return filled


def generate_voxel_grid(bbox, szval, use_cubes=False):
    """Generates the voxel grid with the desired size.

    :param bbox: bounding box
    :type bbox: list, tuple
    :param szval: size in x-, y-, z-directions
    :type szval: list, tuple
    :param use_cubes: use cube voxels instead of cuboid ones
    :type use_cubes: bool
    :return: voxel grid
    :rtype: list
    """
    # Input validation
    if szval[0] <= 1 or szval[1] <= 1 or szval[2] <= 1:
        raise GeomdlException("Size values must be bigger than 1", data=dict(sizevals=szval))

    # Find step size for each direction
    steps = [float(bbox[1][idx] - bbox[0][idx]) / float(szval[idx] - 1) for idx in range(0, 3)]

    # It is possible to use cubes instead of cuboids
    if use_cubes:
        min_val = min(*steps)
        steps = [min_val for _ in range(0, 3)]

    # Find range in each direction
    ranges = [list(linalg.frange(bbox[0][idx], bbox[1][idx], steps[idx])) for idx in range(0, 3)]

    voxel_grid = []
    for u in ranges[0]:
        for v in ranges[1]:
            for w in ranges[2]:
                bbmin = [u, v, w]
                bbmax = [k + l for k, l in zip(bbmin, steps)]
                voxel_grid.append([bbmin, bbmax])
    return voxel_grid


def is_point_inside_voxel(bbox, ptsarr, **kwargs):
    """Finds if any point is contained inside the voxel boundaries (inouts array).

    Ref: https://math.stackexchange.com/a/1552579

    :param bbox: bounding box of the voxel
    :type bbox: list, tuple
    :param ptsarr: points to be checked
    :type ptsarr: list, tuple
    :return: list of ins and outs
    :rtype: list
    """
    # Get keyword arguments
    tol = kwargs.get("tol", 10e-8)  # padding value

    # Make bounding box vertices more readable
    bbmin = [b - tol for b in bbox[0]]
    bbmax = [b + tol for b in bbox[1]]

    # Find basis vectors
    i = [bbmax[0] - bbmin[0], 0, 0]
    j = [0, bbmax[1] - bbmin[1], 0]
    k = [0, 0, bbmax[2] - bbmin[2]]

    # Find dot products
    idi = linalg.vector_dot(i, i)
    jdj = linalg.vector_dot(j, j)
    kdk = linalg.vector_dot(k, k)

    for pt in ptsarr:
        v = [p - b for p, b in zip(pt, bbmin)]
        # Bigger than and equal to will include the border and,
        # since we have a padding on the boundary box, we only
        # need to include the lower boundary below
        vdi = linalg.vector_dot(v, i)
        vdj = linalg.vector_dot(v, j)
        vdk = linalg.vector_dot(v, k)
        if idi > vdi >= 0.0 and jdj > vdj >= 0.0 and kdk > vdk >= 0.0:
            return 1
    return 0


def get_points_inside_voxel(bbox, ptsarr, **kwargs):
    """Finds the list of points contained inside the voxel boundaries.

    Ref: https://math.stackexchange.com/a/1552579

    :param bbox: bounding box of the voxel
    :type bbox: list, tuple
    :param ptsarr: points to be checked
    :type ptsarr: list, tuple
    :return: list of points inside the voxel
    :rtype: list
    """
    # Get keyword arguments
    tol = kwargs.get("tol", 10e-8)

    # Make bounding box vertices more readable
    bbmin = [b - tol for b in bbox[0]]
    bbmax = [b + tol for b in bbox[1]]

    # Initialize an empty list
    points_inside = []

    # Find basis vectors
    i = [bbmax[0] - bbmin[0], 0, 0]
    j = [0, bbmax[1] - bbmin[1], 0]
    k = [0, 0, bbmax[2] - bbmin[2]]

    # Find dot products
    idi = linalg.vector_dot(i, i)
    jdj = linalg.vector_dot(j, j)
    kdk = linalg.vector_dot(k, k)

    for pt in ptsarr:
        v = [p - b for p, b in zip(pt, bbmin)]
        # Bigger than and equal to will include the border and,
        # since we have a padding on the boundary box, we only
        # need to include the lower boundary below
        vdi = linalg.vector_dot(v, i)
        vdj = linalg.vector_dot(v, j)
        vdk = linalg.vector_dot(v, k)
        if idi > vdi >= 0.0 and jdj > vdj >= 0.0 and kdk > vdk >= 0.0:
            points_inside.append(pt)
    return points_inside



"""
.. module:: freeform
    :platform: Unix, Windows
    :synopsis: Provides freeform geometry classes

.. moduleauthor:: Onur Rauf Bingol <orbingol@gmail.com>

"""

from nurbs import abstract


class Freeform(abstract.Geometry):
    """n-dimensional freeform geometry"""

    def __init__(self, **kwargs):
        super(Freeform, self).__init__(**kwargs)
        self._geometry_type = "freeform"
        self.name = "freeform geometry"

    @property
    def data(self):
        """Returns a dict which contains the geometry data.

        Please refer to the `wiki <https://github.com/orbingol/NURBS-Python/wiki/Using-Python-Properties>`_ for details
        on using this class member.
        """
        return dict(type=self.type, points=tuple(self.evalpts))

    def evaluate(self, **kwargs):
        """Sets points that form the geometry.

        Keyword Arguments:
            * ``points``: sets the points
        """
        self._eval_points = kwargs.get("points", self._init_array())
        self._dimension = len(self._eval_points[0])



""" Object-oriented B-Spline and NURBS evaluation library in pure Python

.. moduleauthor:: Onur Rauf Bingol <orbingol@gmail.com>

"""

# Library version
__version__ = "0.0.1"

# Author and license
__author__ = "Onur Rauf Bingol & Dessia Technologies"
__license__ = "MIT"

# Optional variables
__description__ = "Object-oriented B-Spline and NURBS evaluation library"
__keywords__ = "NURBS B-Spline curve surface CAD modeling visualization surface-generator"

# Support for "from nurbs import *"
# @see: https://stackoverflow.com/a/41895257
# @see: https://stackoverflow.com/a/35710527
__all__ = [
    "BSpline",
    "compatibility",
    "construct",
    "convert",
    "CPGen",
    "elements",
    "evaluators",
    "exchange",
    "exchange_vtk",
    "fitting",
    "helpers",
    "linalg",
    "multi",
    "NURBS",
    "operations",
    "ray",
    "tessellate",
    "utilities",
    "voxelize",
]


def geomdl_version():
    """Returns nurbs full version as a tuple"""
    return tuple(__version__.split("."))



#!/usr/bin/env python

import os
from os.path import dirname, isdir, join
import re
import shutil
import sys
from distutils.command.clean import clean as clean_command
from subprocess import CalledProcessError, check_output

from setuptools import Extension, setup
from setuptools.command.install import install as install_command
from setuptools.command.test import test as test_command

# Global variables to control generation of optional Cython-compiled library core module
BUILD_FROM_CYTHON = False
BUILD_FROM_SOURCE = False


def read(file_name):
    return open(os.path.join(os.path.dirname(__file__), file_name)).read()


# Implemented from http://stackoverflow.com/a/41110107
def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open(project + "/__init__.py").read())
    return result.group(1)


class InstallCommand(install_command):
    """Overrides pip install command to control generation of optional Cython-compiled library core module"""

    user_options = install_command.user_options + [
        ("use-cython", None, "Cythonize and compile nurbs.core"),
        ("use-source", None, "Compile nurbs.core from the source files"),
        ("core-only", None, "Compile and install nurbs.core only"),
    ]

    def initialize_options(self):
        install_command.initialize_options(self)
        self.use_cython = 0
        self.use_source = 0

    def finalize_options(self):
        install_command.finalize_options(self)

    def run(self):
        global BUILD_FROM_CYTHON, BUILD_FROM_SOURCE
        BUILD_FROM_CYTHON = True if self.use_cython > 0 else False
        BUILD_FROM_SOURCE = True if self.use_source > 0 else False
        install_command.run(self)


# Reference: https://docs.pytest.org/en/latest/goodpractices.html
class PyTest(test_command):
    """Allows test command to call py.test"""

    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        test_command.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


class SetuptoolsClean(clean_command):
    """Cleans Cython-generated source files and setuptools-generated directories"""

    def run(self):
        # Call parent method
        clean_command.run(self)

        # Clean setuptools-generated directories
        st_dirs = ["dist", "build", "nurbs.egg-info", "nurbs.core.egg-info"]

        print("Removing setuptools-generated directories")
        for d in st_dirs:
            d_path = os.path.join(os.path.dirname(__file__), d)
            shutil.rmtree(d_path, ignore_errors=True)

        # Find list of files with .c extension
        flist_c, flist_c_path = read_files("nurbs", ".c")

        # Clean files with .c extensions
        if flist_c_path:
            print("Removing Cython-generated source files with .c extension")
            for f in flist_c_path:
                f_path = os.path.join(os.path.dirname(__file__), f)
                os.unlink(f_path)

        # Find list of files with .cpp extension
        flist_cpp, flist_cpp_path = read_files("nurbs", ".cpp")

        # Clean files with .cpp extensions
        if flist_cpp_path:
            print("Removing Cython-generated source files with .cpp extension")
            for f in flist_cpp_path:
                f_path = os.path.join(os.path.dirname(__file__), f)
                os.unlink(f_path)


def read_files(project, ext):
    """Reads files inside the input project directory."""
    project_path = os.path.join(os.path.dirname(__file__), project)
    file_list = os.listdir(project_path)
    flist = []
    flist_path = []
    for f in file_list:
        f_path = os.path.join(project_path, f)
        if os.path.isfile(f_path) and f.endswith(ext) and f != "__init__.py":
            flist.append(f.split(".")[0])
            flist_path.append(f_path)
    return flist, flist_path


def copy_files(src, ext, dst):
    """Copies files with extensions "ext" from "src" to "dst" directory."""
    src_path = os.path.join(os.path.dirname(__file__), src)
    dst_path = os.path.join(os.path.dirname(__file__), dst)
    file_list = os.listdir(src_path)
    for f in file_list:
        if f == "__init__.py":
            continue
        f_path = os.path.join(src_path, f)
        if os.path.isfile(f_path) and f.endswith(ext):
            shutil.copy(f_path, dst_path)


def make_dir(project):
    """Creates the project directory for compiled modules."""
    project_path = os.path.join(os.path.dirname(__file__), project)
    # Delete the directory and the files inside it
    if os.path.exists(project_path):
        shutil.rmtree(project_path)
    # Create the directory
    os.mkdir(project_path)
    # We need a __init__.py file inside the directory
    with open(os.path.join(project_path, "__init__.py"), "w") as fp:
        fp.write('__version__ = "' + str(get_property("__version__", "nurbs")) + '"\n')
        fp.write('__author__ = "' + str(get_property("__author__", "nurbs")) + '"\n')
        fp.write('__license__ = "' + str(get_property("__license__", "nurbs")) + '"\n')


def in_argv(arg_list):
    """Checks if any of the elements of the input list is in sys.argv array."""
    for arg in sys.argv:
        for parg in arg_list:
            if parg == arg or arg.startswith(parg):
                return True
    return False


def readme():
    with open("README.md") as f:
        return f.read()


tag_re = re.compile(r"\btag: %s([0-9][^,]*)\b")
version_re = re.compile("^Version: (.+)$", re.M)


def version_from_git_describe(version):
    if version[0] == "v":
        version = version[1:]
    # PEP 440 compatibility
    number_commits_ahead = 0
    if "-" in version:
        version, number_commits_ahead, commit_hash = version.split("-")
        number_commits_ahead = int(number_commits_ahead)

    split_versions = version.split(".")
    if "post" in split_versions[-1]:
        suffix = split_versions[-1]
        split_versions = split_versions[:-1]
    else:
        suffix = None
    for pre_release_segment in ["a", "b", "rc"]:
        if pre_release_segment in split_versions[-1]:
            if number_commits_ahead > 0:
                split_versions[-1] = str(split_versions[-1].split(pre_release_segment)[0])
                if len(split_versions) == 2:
                    split_versions.append("0")
                if len(split_versions) == 1:
                    split_versions.extend(["0", "0"])
                split_versions[-1] = str(int(split_versions[-1]) + 1)
                future_version = ".".join(split_versions)
                return "{}.dev{}+{}".format(future_version, number_commits_ahead, commit_hash)
            else:
                return ".".join(split_versions)
    if number_commits_ahead > 0:
        if len(split_versions) == 2:
            split_versions.append("0")
        if len(split_versions) == 1:
            split_versions.extend(["0", "0"])
        split_versions[-1] = str(int(split_versions[-1]) + 1)
        split_versions = ".".join(split_versions)
        return "{}.dev{}+{}".format(split_versions, number_commits_ahead, commit_hash)
    else:
        if suffix is not None:
            split_versions.append(suffix)
        return ".".join(split_versions)


# Just testing if get_version works well
assert version_from_git_describe("v0.1.7.post2") == "0.1.7.post2"
assert version_from_git_describe("v0.0.1-25-gaf0bf53") == "0.0.2.dev25+gaf0bf53"
assert version_from_git_describe("v0.1-15-zsdgaz") == "0.1.1.dev15+zsdgaz"
assert version_from_git_describe("v1") == "1"
assert version_from_git_describe("v1-3-aqsfjbo") == "1.0.1.dev3+aqsfjbo"
assert version_from_git_describe("v0.13rc0") == "0.13rc0"


def get_version():
    # Return the version if it has been injected into the file by git-archive
    version = tag_re.search("$Format:%D$")
    if version:
        return version.group(1)
    d = dirname(__file__)

    if isdir(join(d, ".git")):
        cmd = "git describe --tags"
        try:
            version = check_output(cmd.split()).decode().strip()[:]
        except CalledProcessError:
            raise RuntimeError("Unable to get version number from git tags")
        version = version_from_git_describe(version)
    else:
        # Extract the version from the PKG-INFO file.
        with open(join(d, "PKG-INFO")) as f:
            version = version_re.search(f.read()).group(1)

    # branch = get_branch()
    # if branch and branch != 'master':
    #     branch = re.sub('[^A-Za-z0-9]+', '', branch)
    #     if '+' in version:
    #         version += f'{branch}'
    #     else:
    #         version += f'+{branch}'

    return version


def get_branch():
    if isdir(join(dirname(__file__), ".git")):
        cmd = "git branch --show-current"
        try:
            return check_output(cmd.split()).decode().strip()[:]
        except CalledProcessError:
            pass

    return None


# Define setup.py commands to activate Cython compilation
possible_cmds = ["install", "build", "bdist"]

# Use geomdl.core package only
if "--core-only" in sys.argv:
    package_name = "nurbs.core"
    package_dir = "nurbs/core"
    packages = []
    sys.argv.remove("--core-only")
    sys.argv.append("--use-cython")
else:
    package_name = package_dir = "nurbs"
    packages = ["nurbs", "nurbs.visualization"]

# geomdl.core compilation
# Ref: https://gist.github.com/ctokheim/6c34dc1d672afca0676a

# Use already Cythonized C code
if in_argv(possible_cmds) and "--use-source" in sys.argv:
    BUILD_FROM_SOURCE = True
    sys.argv.remove("--use-source")

# Use Cython to (re)generate C code (overrides "--use-source")
if in_argv(possible_cmds) and "--use-cython" in sys.argv:
    # Try to import Cython
    try:
        from Cython.Build import cythonize
    except ImportError:
        raise ImportError("Cython is required for this step. Please install it via 'pip install cython'")

    BUILD_FROM_CYTHON = True
    BUILD_FROM_SOURCE = False
    sys.argv.remove("--use-cython")

# We don't want to include any compiled files with the distribution
ext_modules = []

if BUILD_FROM_CYTHON or BUILD_FROM_SOURCE:
    # Choose the file extension
    file_ext = ".py" if BUILD_FROM_CYTHON else ".c"

    # Create Cython-compiled module directory
    make_dir("nurbs/core")

    # Create extensions
    optional_extensions = []
    fnames, fnames_path = read_files("nurbs", file_ext)
    for fname, fpath in zip(fnames, fnames_path):
        temp = Extension("nurbs.core." + str(fname), sources=[fpath])
        optional_extensions.append(temp)

    # Call Cython when "python setup.py build_ext --use-cython" is executed
    if BUILD_FROM_CYTHON:
        ext_modules = cythonize(optional_extensions, compiler_directives={"language_level": sys.version_info[0]})

    # Compile from C source when "python setup.py build_ext --use-source" is executed
    if BUILD_FROM_SOURCE:
        ext_modules = optional_extensions

    # Add Cython-compiled module to the packages list
    packages.append("nurbs.core")

# Input for setuptools.setup
data = dict(
    name=package_name,
    version=get_version(),
    description=get_property("__description__", package_dir),
    long_description=read("DESCRIPTION.rst"),
    license=get_property("__license__", package_dir),
    author=get_property("__author__", package_dir),
    author_email="root@dessia.tech",
    url="https://github.com/Dessia-tech/nurbs",
    keywords=get_property("__keywords__", package_dir),
    packages=packages,
    install_requires=[],
    tests_require=["pytest>=3.6.0"],
    cmdclass={"install": InstallCommand, "test": PyTest, "clean": SetuptoolsClean},
    ext_modules=ext_modules,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Visualization",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    project_urls={
        "Documentation": "https://github.com/Dessia-tech/nurbs/doc",
        "Source": "https://github.com/Dessia-tech/nurbs",
        "Tracker": "https://github.com/Dessia-tech/nurbs/issues",
    },
)


if __name__ == "__main__":
    setup(**data)
